/*
 * Export MIPS Creator CI20 HDMI display operations as a server.
 *
 * Copyright (C) 2018, 2020, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hdmi-jz4780.h>
#include <l4/devices/lcd-jz4740-config.h>
#include <l4/devices/lcd-jz4740-panel.h>
#include <l4/devices/panel-loader.h>
#include <l4/devices/memory.h>

#include <l4/re/env.h>

#include <ipc/irq.h>
#include <ipc/server.h>
#include "activation_server.h"
#include "cpm_client.h"
#include "icu_client.h"

/* Virtual addresses for the HDMI register block plus interrupt capability. */

static l4_addr_t hdmi_virt_base = 0, hdmi_virt_base_end = 0;
static l4_uint32_t hdmi_irq_start, hdmi_irq_end;
static l4_cap_idx_t hdmi_irq, icu;

/* Device abstractions. */

static Hdmi_jz4780_chip *hdmi = 0;
static CPM *cpm_device;



static int setup_memory(void)
{
  if (get_memory("jz4780-hdmi", &hdmi_virt_base, &hdmi_virt_base_end))
    return 1;

  if (get_irq("jz4780-hdmi", &hdmi_irq_start, &hdmi_irq_end))
    return 1;

  /* Obtain a reference to the CPM device. */

  l4_cap_idx_t cpm = l4re_env_get_cap("cpm");
  if (!l4_is_valid_cap(cpm)) return 1;

  static client_CPM cpm_obj(cpm);
  cpm_device = &cpm_obj;

  /* Start the HDMI peripheral. */

  cpm_device->stop_clock(Clock_hdmi);
  cpm_device->set_frequency(Clock_frequency_hdmi, 27000000);
  cpm_device->start_clock(Clock_hdmi);

  /* Load the panel data from the configured library. */

  struct Jz4740_lcd_panel *panel = (struct Jz4740_lcd_panel *) load_panel();
  if (!panel) return 1;

  /* Obtain access to the HDMI interrupt using the ICU. */

  icu = l4re_env_get_cap("icu");
  if (!l4_is_valid_cap(icu)) return 1;

  if (ipc_create_irq(&hdmi_irq)) return 1;

  /* Bind the IRQ object to the interrupt in this thread. */

  client_ICU icu_obj(icu);

  long err = icu_obj.bind(hdmi_irq_start, hdmi_irq);
  if (err) return err;

  err = ipc_bind_irq(hdmi_irq, 0, l4re_env()->main_thread);
  if (err) return err;

  /* Obtain an abstraction for the HDMI peripheral. */

  hdmi = new Hdmi_jz4780_chip(hdmi_virt_base, hdmi_virt_base_end, hdmi_irq, panel);

  return 0;
}



/* Display device. */

class server_DisplayObject : public Activation
{
public:
  /* Switch the display off. */

  long disable(void)
  {
    return L4_EOK;
  }

  /* Switch the display on. */

  long enable(void)
  {
    uint32_t frequency;

    // NOTE: Should test to see if it is connected.

    if (!cpm_device->get_frequency(Clock_frequency_lcd_pixel, &frequency))
    {
      hdmi->enable(frequency);
    }

    return L4_EOK;
  }
};



int main(void)
{
  if (setup_memory()) return 1;

  /* Initialise and register a new server object. */

  server_DisplayObject obj;
  l4_cap_idx_t server;

  if (ipc_server_bind("display", (l4_umword_t) &obj, &server)) return 1;

  /* Enter the IPC server loop. */

  ipc_server_loop(Activation_expected_items, &obj,
                  (ipc_server_handler_type) handle_Activation);
  return 0;
}
