/*
 * Export a JZ4730 PWM peripheral as a server.
 *
 * Copyright (C) 2018, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/gpio-jz4730.h>
#include <l4/devices/pwm-jz4730.h>
#include <l4/devices/memory.h>

#include <l4/re/env.h>

#include <stdint.h>
#include <stdlib.h>

#include <ipc/server.h>
#include "pwm_object_server.h"

/* Virtual addresses for the GPIO and PWM register blocks. */

static l4_addr_t gpio_virt_base = 0, gpio_virt_base_end = 0;
static l4_addr_t pwm_virt_base = 0, pwm_virt_base_end = 0;



static int setup_memory(void)
{
  if (get_memory("jz4730-gpio", &gpio_virt_base, &gpio_virt_base_end))
    return 1;

  if (get_memory("jz4730-pwm", &pwm_virt_base, &pwm_virt_base_end))
    return 1;

  return 0;
}



/* PWM peripheral device. */

class server_PWMObject : public PWMObject
{
  Pwm_jz4730_chip *_device = 0;
  int _duty, _period, _prescale;

public:
  /* Associate the device with a particular memory region. */

  explicit server_PWMObject(Pwm_jz4730_chip *device, int duty, int period,
                            int prescale)
  : _device(device), _duty(duty), _period(period), _prescale(prescale)
  {
  }

  /* Disable the device. */

  long disable()
  {
    _device->disable();
    return L4_EOK;
  }

  /* Enable the device. */

  long enable()
  {
    set_duty(_duty);
    set_period(_period);
    set_control(0x80 | _prescale); /* enable | prescale */
    return L4_EOK;
  }

  /* Set the control register. */

  long set_control(uint8_t control)
  {
    _device->set_control(control);
    return L4_EOK;
  }

  /* Set the PWM duty cycle. */

  long set_duty(uint16_t duty)
  {
    _duty = duty;
    _device->set_duty(duty);
    return L4_EOK;
  }

  /* Set the PWM period. */

  long set_period(uint16_t period)
  {
    _period = period;
    _device->set_period(period);
    return L4_EOK;
  }
};



/* Initialise devices and start the server. */

static int run(int number, int duty, int period, int prescale)
{
  Pwm_jz4730_chip pwm_device(pwm_virt_base + number * 0x1000,
                             pwm_virt_base + (number + 1) * 0x1000);

  Gpio_jz4730_chip gpio_port_c(gpio_virt_base + 2 * 0x30,
                               gpio_virt_base + 3 * 0x30, 32);

  /* Enable the PWM output for PC30 or PC31. */

  gpio_port_c.config_pad(30 + number, Hw::Gpio_chip::Function_alt, 1);

  /* Initialise and register a new server object. */

  server_PWMObject obj(&pwm_device, duty, period, prescale);
  l4_cap_idx_t server;

  if (ipc_server_bind("pwm", (l4_umword_t) &obj, &server)) return 1;

  /* Enter the IPC server loop. */

  ipc_server_loop(PWMObject_expected_items, &obj,
                  (ipc_server_handler_type) handle_PWMObject);
  return 0;
}



/* Arguments: <PWM device number> <duty> <period> <prescale> */

int main(int argc, char *argv[])
{
  int number, duty, period, prescale;

  if (argc < 5) return 1;

  /* Interpret and restrict the device number. */

  number = atoi(argv[1]);
  if ((number < 0) || (number > 1)) return 1;

  /* Interpret and restrict the initial parameters. */

  duty = atoi(argv[2]);
  period = atoi(argv[3]);
  prescale = atoi(argv[4]);
  if ((duty < 0) || (period < 0) || (prescale < 0)) return 1;

  /* Obtain access to peripheral memory. */

  if (setup_memory()) return 1;

  return run(number, duty, period, prescale);
}
